/*
 *      ccard - credit card number validation
 *      Copyright (C) 2007, 2009 Peter Miller
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see
 *      <http://www.gnu.org/licenses/>.
 *
 * derived from code by
 *      Chris Stone <cstone@hms.com>
 *      The High Mountain Software Internet Gateway
 *
 * translated to C by
 *      Peter Miller, 1994-Oct-28
 */

#include <ctype.h>
#include <lib/ac/string.h>

#include <ccard/ccard.h>

#define SIZEOF(a) (sizeof(a) / sizeof((a)[0]))
#define ENDOF(a) ((a) + SIZEOF(a))

#define MINLEN 12
#define MAXLEN (MINLEN + 16)
#define L(n) (1 << ((n) - MINLEN))


const char *
ccard_type_name(ccard_type_ty n)
{
    static char *name[] =
    {
        "unknown",
        "Mastercard",
        "Visa",
        "American Express",
        "Diners Club/Carte Blanche",
        "Discover",
        "enRoute",
        "Japanese Credit Bureau",
        "Australian Bankcard",
    };

    if (n >= SIZEOF(name))
        n = 0;
    return name[n];
}


const char *
ccard_error_name(ccard_error_ty n)
{
    static char *name[] =
    {
        "no error",
        "card type unknown",
        "card number contains non numeric characters",
        "card number is far too long",
        "card number is the wrong length",
        "checksum incorrect",
    };

    if (n >= SIZEOF(name))
        return "unknown";
    return name[n];
}


static int
verify_checksum(const char *credit_card)
{
    const char      *cp;
    int             dbl;
    int             check_sum;

    /*
     * This is the Luhn error check algorithm.
     *
     * http://en.wikipedia.org/wiki/Luhn_algorithm
     */
    check_sum = 0;
    dbl = 0;
    /* assert(credit_card[0]); */
    cp = credit_card + strlen(credit_card) - 1;
    while (cp >= credit_card)
    {
        int             c;

        c = *cp-- - '0';
        if (dbl)
        {
            c *= 2;
            if (c >= 10)
                c -= 9;
        }
        check_sum += c;
        dbl = !dbl;
    }

    return ((check_sum % 10) == 0);
}


static int
all_numeric(const char *s1, char *s2, int max)
{
    while (*s1)
    {
        if (isspace((unsigned char)*s1) || *s1 == '-')
        {
            ++s1;
            continue;
        }
        if (!isdigit(*s1))
            return ccard_error_non_numeric;
        if (max <= 0)
            return ccard_error_too_long;
        *s2++ = *s1++;
        --max;
    }
    *s2 = 0;
    return ccard_error_none;
}


ccard_error_ty
ccard_valid(const char *credit_card_in, ccard_type_ty *card_type)
{
    typedef struct table_ty table_ty;
    struct table_ty
    {
        char            *prefix;
        int             length_mask;
        ccard_type_ty   type;
        int             checksum;
    };

    static table_ty table[] =
    {
        { "1800", L(15), ccard_type_jcb, 1 },
        { "2014", L(15), ccard_type_enroute, 0 },
        { "2131", L(15), ccard_type_jcb, 1 },
        { "2149", L(15), ccard_type_enroute, 0 },
        { "300", L(14), ccard_type_diners, 1 },
        { "301", L(14), ccard_type_diners, 1 },
        { "302", L(14), ccard_type_diners, 1 },
        { "303", L(14), ccard_type_diners, 1 },
        { "304", L(14), ccard_type_diners, 1 },
        { "305", L(14), ccard_type_diners, 1 },
        { "34", L(15), ccard_type_amex, 1 },
        { "36", L(14), ccard_type_diners, 1 },
        { "37", L(15), ccard_type_amex, 1 },
        { "38", L(14), ccard_type_diners, 1 },
        { "3", L(16), ccard_type_jcb, 1 },
        { "4", L(13) | L(16), ccard_type_visa, 1 },
        { "51", L(16), ccard_type_mastercard, 1 },
        { "52", L(16), ccard_type_mastercard, 1 },
        { "53", L(16), ccard_type_mastercard, 1 },
        { "54", L(16), ccard_type_mastercard, 1 },
        { "55", L(16), ccard_type_mastercard, 1 },
        { "56", L(16), ccard_type_bankcard, 1 },
        { "6011", L(16), ccard_type_discover, 1 },
    };
    table_ty        *tp;
    char            credit_card[MAXLEN + 1];
    ccard_error_ty  err;
    int             len;

    /*
     * copy the number, eliding spaces
     * defer any errors until after we have tried to guess the card type
     */
    err = all_numeric(credit_card_in, credit_card, MAXLEN);

    /*
     * look for the card prefix in the table
     * to determine the card type
     */
    for (tp = table; tp < ENDOF(table); ++tp)
    {
        if (!memcmp(tp->prefix, credit_card, strlen(tp->prefix)))
            break;
    }
    if (tp >= ENDOF(table))
    {
        *card_type = ccard_type_unknown;
        return ccard_error_type_unknown;
    }
    *card_type = tp->type;
    if (err != ccard_error_none)
        return err;

    /*
     * set the card type, then check the length
     */
    /* assert(tp->correct_length <= MAXLEN); */
    len = strlen(credit_card);
    if (len < MINLEN || (L(len) & tp->length_mask) == 0)
        return ccard_error_length_incorrect;

    /*
     * checksum
     */
    if (tp->checksum && !verify_checksum(credit_card))
        return ccard_error_checksum;

    /*
     * no errors found
     */
    return ccard_error_none;
}


const char *
ccard_industry_identifier(const char *name)
{
    static const char *table[10] =
    {
        "Other",
        "Airlines",
        "Airlines",
        "Travel/Entertainment",
        "Banking/Financial",
        "Banking/Financial",
        "Merchandising/Financial",
        "Petroleum",
        "Telecommunications",
        "Other",
    };

    for (;;)
    {
        unsigned char   c;

        c = *name++;
        if (!c)
            return "Unknown";
        if (isdigit(c))
            return table[c - '0'];
    }
}
